import json
from argparse import ArgumentParser
from datetime import datetime
from functools import reduce
from pprint import pprint
from typing import Any, Dict, List

from diagnnose.utils.misc import merge_dicts

from .arg_descriptions import arg_descriptions

# group -> arg_name -> value
ConfigDict = Dict[str, Dict[str, Any]]


def create_config_dict() -> ConfigDict:
    """Sets up the configuration for extraction.

    Config can be provided from a json file or the commandline. Values
    in the json file can be overwritten by providing them from the
    commandline. Will raise an error if a required argument is not
    provided in either the json file or as a commandline arg.

    Commandline args should be provided as dot-separated values, where
    the first dot indicates the arg group the arg belongs to.
    For example, setting the ``state_dict`` of a ``model`` can be done
    with the flag ``--model.state_dict state_dict``.

    Returns
    -------
    config_dict : ConfigDict
        Dictionary mapping each arg group to their config values.
    """
    arg_parser = _create_arg_parser()

    args, unk_args = arg_parser.parse_known_args()
    cmd_args = vars(args)

    # Load arguments from config
    init_config_dict: ConfigDict = {}
    if cmd_args["config"] is not None:
        with open(cmd_args.pop("config")) as f:
            init_config_dict.update(json.load(f))

    _add_unk_args(cmd_args, unk_args)
    config_dict = _add_cmd_args(init_config_dict, cmd_args)

    _cast_activation_names(config_dict)

    _set_tokenizer(config_dict)

    print(datetime.now().strftime("%d/%m/%Y %H:%M:%S"))
    pprint(config_dict)

    return config_dict


def _create_arg_parser() -> ArgumentParser:
    parser = ArgumentParser()

    # create group to load config from a file
    from_config = parser.add_argument_group(
        "From config file", "Provide full experiment setup via config json file."
    )
    from_config.add_argument(
        "-c", "--config", help="Path to json file containing extraction config."
    )

    # create group to provide info via commandline arguments
    # Required args are not set to be required here as they can come from --config
    from_cmd = parser.add_argument_group(
        "From commandline",
        "Specify experiment setup via commandline arguments. "
        "Can be combined with the json config, in which case these"
        " cmd arguments overwrite the config args.",
    )

    for group, group_args in arg_descriptions.items():
        for arg, arg_config in group_args.items():
            arg = f"{group}.{arg}"

            if arg_config.get("type", str) == bool:
                action = (
                    "store_false" if arg_config.get("default", False) else "store_true"
                )
                from_cmd.add_argument(
                    f"--{arg}", action=action, help=arg_config["help"], default=None
                )
            else:
                from_cmd.add_argument(
                    f"--{arg}",
                    nargs=arg_config.get("nargs", None),
                    type=arg_config.get("type", str),
                    help=arg_config["help"],
                )

    return parser


def _add_unk_args(cmd_args: Dict[str, Any], unk_args: List[str]):
    """ Add arguments that are not part of the default arg structure """
    unk_args = [x.split() for x in " ".join(unk_args).split("--") if len(x) > 0]
    for arg in unk_args:
        key = arg[0]
        val = arg[1] if len(arg) == 2 else arg[1:]
        cmd_args[key] = val


def _add_cmd_args(config_dict: ConfigDict, cmd_args: Dict[str, Any]) -> ConfigDict:
    """ Update provided config values with cmd args. """
    cdm_arg_dicts: List[ConfigDict] = []
    for arg, val in cmd_args.items():
        if val is not None:
            cmd_arg_keys = arg.split(".")[::-1]
            cmd_arg_dict = val
            for key in cmd_arg_keys:
                cmd_arg_dict = {key: cmd_arg_dict}
            cdm_arg_dicts.append(cmd_arg_dict)

    if len(cdm_arg_dicts) > 0:
        return reduce(merge_dicts, (config_dict, *cdm_arg_dicts))

    return config_dict


def _cast_activation_names(config_dict: ConfigDict) -> None:
    """Casts activation names tobthe tuple format that is used
    throughout the library.
    """
    # Translate activation names to tuple format that is used in the library
    for group, group_config in config_dict.items():
        if "activation_names" in group_config:
            activation_names = group_config["activation_names"]
            assert all(
                isinstance(a_name, list)
                and isinstance(a_name[0], int)
                and isinstance(a_name[1], str)
                for a_name in activation_names
            ), "Incorrect format for activation names, should be [[layer, name]]."

            config_dict[group]["activation_names"] = list(map(tuple, activation_names))


def _set_tokenizer(config_dict: ConfigDict) -> None:
    """ Set tokenizer name manually for Huggingface models. """
    if (
        "transformer_type" in config_dict.get("model", {})
        and "tokenizer" not in config_dict
    ):
        config_dict["tokenizer"] = {"path": config_dict["model"]["transformer_type"]}

        if "cache_dir" in config_dict["model"]:
            config_dict["tokenizer"]["cache_dir"] = config_dict["model"]["cache_dir"]



from typing import Any, Dict

from .forward_lstm import ForwardLSTM


class AWDLSTM(ForwardLSTM):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        kwargs.setdefault("rnn_name", "rnns")
        super().__init__(*args, **kwargs)

    @staticmethod
    def param_names(layer: int, rnn_name: str, **kwargs) -> Dict[str, str]:
        return {
            "weight_hh": f"{rnn_name}.{layer}.module.weight_hh_l0_raw",
            "weight_ih": f"{rnn_name}.{layer}.module.weight_ih_l0",
            "bias_hh": f"{rnn_name}.{layer}.module.bias_hh_l0",
            "bias_ih": f"{rnn_name}.{layer}.module.bias_ih_l0",
        }



from contextlib import ExitStack
from typing import TYPE_CHECKING, Optional, Union

import torch
from torchtext.data import Batch
from tqdm import tqdm

import diagnnose.activations.selection_funcs as selection_funcs
from diagnnose.activations import ActivationReader, ActivationWriter
from diagnnose.activations.selection_funcs import return_all
from diagnnose.corpus import Corpus
from diagnnose.corpus.create_iterator import create_iterator
from diagnnose.typedefs.activations import (
    ActivationDict,
    ActivationNames,
    ActivationRanges,
    SelectionFunc,
)

if TYPE_CHECKING:
    from diagnnose.models import LanguageModel

BATCH_SIZE = 1024


class Extractor:
    """Extracts all intermediate activations of a LM from a corpus.

    Only activations that are provided in activation_names will be
    stored in a pickle file. Each activation is written to its own file.

    Parameters
    ----------
    model : LanguageModel
        Language model that inherits from LanguageModel.
    corpus : Corpus
        Corpus containing sentences to be extracted.
    activation_names : List[tuple[int, str]], optional
        List of (layer, activation_name) tuples. If not provided all
        activation_names corresponding to the ``model`` will be
        extracted.
    activations_dir : str, optional
        Directory to which activations will be written. If not provided
        the `extract()` method will only return the activations without
        writing them to disk.
    selection_func : Union[SelectionFunc, str]
        Function which determines if activations for a token should
        be extracted or not. Can also be provided as a string,
        indicating the method name of one of the default
        selection_funcs in
        :py:mod:`diagnnose.activations.selection_funcs`.
    batch_size : int, optional
        Amount of sentences processed per forward step. Higher batch
        size increases extraction speed, but should be done
        accordingly to the amount of available RAM. Defaults to 1.
    """

    def __init__(
        self,
        model: "LanguageModel",
        corpus: Corpus,
        activation_names: Optional[ActivationNames] = None,
        activations_dir: Optional[str] = None,
        selection_func: Union[SelectionFunc, str] = return_all,
        batch_size: int = BATCH_SIZE,
    ) -> None:
        self.model = model
        self.corpus = corpus
        self.activation_names = activation_names or model.activation_names()
        if isinstance(selection_func, str):
            self.selection_func = getattr(selection_funcs, selection_func)
        else:
            self.selection_func = selection_func
        self.batch_size = batch_size

        self.activation_ranges = []
        self.set_activation_ranges()

        if activations_dir is None:
            self.activation_writer: Optional[ActivationWriter] = None
        else:
            self.activation_writer = ActivationWriter(activations_dir)

    def extract(self) -> ActivationReader:
        """Extracts embeddings from a corpus.

        Uses :class:`contextlib.ExitStack` to write to multiple files
        simultaneously.

        Returns
        -------
        activation_reader : ActivationReader
            After extraction an activation_reader is returned that
            provides direct access to the extracted activations.
        """
        print(f"\nStarting extraction of {len(self.corpus)} sentences...")

        if self.activation_writer is not None:
            with ExitStack() as stack:
                self.activation_writer.create_output_files(stack, self.activation_names)

                self._extract_corpus(dump=True)

                self.activation_writer.dump_meta_info(
                    self.activation_ranges, self.selection_func
                )

            activation_reader = ActivationReader(
                activations_dir=self.activation_writer.activations_dir,
                activation_names=self.activation_names,
            )
        else:
            corpus_activations = self._extract_corpus(dump=False)

            activation_reader = ActivationReader(
                activation_dict=corpus_activations,
                activation_names=self.activation_names,
                activation_ranges=self.activation_ranges,
                selection_func=self.selection_func,
            )

        n_extracted = self.activation_ranges[-1][-1]
        print(f"Extraction finished, {n_extracted} activations have been extracted.")

        return activation_reader

    def _extract_corpus(self, dump: bool = True) -> ActivationDict:
        tot_extracted = self.activation_ranges[-1][1]
        corpus_activations: ActivationDict = self._init_activation_dict(
            tot_extracted, dump=dump
        )

        corpus = self._filter_corpus()

        iterator = create_iterator(
            corpus, batch_size=self.batch_size, device=self.model.device
        )

        for batch in tqdm(iterator, unit="batch"):
            batch_activations = self._extract_batch(batch)

            if dump:
                self.activation_writer.dump_activations(batch_activations)
            else:
                # Insert extracted batch activations into full corpus activations dict.
                batch_start = self.activation_ranges[batch.sen_idx[0]][0]
                batch_stop = self.activation_ranges[batch.sen_idx[-1]][1]
                for a_name, activations in batch_activations.items():
                    corpus_activations[a_name][batch_start:batch_stop] = activations

        return corpus_activations

    def _filter_corpus(self) -> Corpus:
        """ Skip items for which selection_func yields 0 activations. """
        sen_ids = [
            idx
            for idx, (start, stop) in enumerate(self.activation_ranges)
            if start != stop
        ]

        if len(sen_ids) != len(self.corpus):
            return self.corpus.slice(sen_ids)

        return self.corpus

    def _extract_batch(self, batch: Batch) -> ActivationDict:
        """Processes the items in `batch` and selects the activations
        that should should be extracted according to selection_func.
        """
        sens, sen_lens = getattr(batch, self.corpus.sen_column)

        compute_out = any("out" in a_name for a_name in self.activation_names)
        with torch.no_grad():
            # a_name -> batch_size x max_sen_len x nhid
            all_activations: ActivationDict = self.model(
                input_ids=sens,
                input_lengths=sen_lens,
                compute_out=compute_out,
                only_return_top_embs=False,
            )

        # a_name -> n_items_in_batch x nhid
        batch_activations = self._select_activations(all_activations, batch)

        return batch_activations

    def _select_activations(
        self,
        all_activations: ActivationDict,
        batch: Batch,
    ) -> ActivationDict:
        """ Selects only the activations that pass selection_func. """
        batch_start = self.activation_ranges[batch.sen_idx[0]][0]
        batch_end = self.activation_ranges[batch.sen_idx[-1]][1]
        n_items_in_batch = batch_end - batch_start
        batch_activations: ActivationDict = self._init_activation_dict(n_items_in_batch)

        a_idx = 0

        for b_idx, sen_idx in enumerate(batch.sen_idx):
            item = self.corpus[sen_idx]
            sen_len = len(getattr(item, self.corpus.sen_column))
            for w_idx in range(sen_len):
                if self.selection_func(w_idx, item):
                    for a_name in batch_activations:
                        selected_activation = all_activations[a_name][b_idx][w_idx]
                        batch_activations[a_name][a_idx] = selected_activation
                    a_idx += 1

        return batch_activations

    def set_activation_ranges(self) -> ActivationRanges:
        activation_ranges: ActivationRanges = []
        tot_extracted = 0

        for item in self.corpus:
            start = tot_extracted
            sen_len = len(getattr(item, self.corpus.sen_column))
            for w_idx in range(sen_len):
                if self.selection_func(w_idx, item):
                    tot_extracted += 1

            activation_ranges.append((start, tot_extracted))

        self.activation_ranges = activation_ranges

    def _init_activation_dict(self, n_items: int, dump: bool = False) -> ActivationDict:
        # If activations are dumped we don't keep track of the full activation dictionary,
        # so an empty dictionary is returned.
        if dump:
            return {}

        corpus_activations = {
            a_name: torch.zeros(
                n_items, self.model.nhid(a_name), device=self.model.device
            )
            for a_name in self.activation_names
        }

        return corpus_activations



from typing import Optional

import torch.nn as nn
import torch.nn.functional as F
from skorch import NeuralNetClassifier
from torch import Tensor


class LogRegModule(nn.Module):
    def __init__(self, ninp: int, nout: int, rank: Optional[int] = None):
        super().__init__()

        if rank is None:
            self.classifier = nn.Linear(ninp, nout)
        else:
            self.classifier = nn.Sequential(
                nn.Linear(ninp, rank), nn.Linear(rank, nout)
            )

    def forward(self, inp: Tensor, create_softmax=True):
        if create_softmax:
            return F.softmax(self.classifier(inp), dim=-1)
        return self.classifier(inp)


# https://github.com/skorch-dev/skorch/blob/master/docs/user/neuralnet.rst#subclassing-neuralnet
class L1NeuralNetClassifier(NeuralNetClassifier):
    def __init__(self, *args, lambda1=0.01, **kwargs):
        super().__init__(*args, **kwargs)
        self.lambda1 = lambda1

    def get_loss(self, y_pred, y_true, X=None, training=False):
        loss = super().get_loss(y_pred, y_true, X=X, training=training)
        loss += self.lambda1 * sum([w.abs().sum() for w in self.module_.parameters()])
        return loss



import contextlib
import cProfile
import io
import pstats
from collections import MutableMapping
from functools import wraps
from typing import Any, Callable, Dict


def suppress_print(func: Callable) -> Callable:
    """
    Function decorator to suppress print output for testing purposes.

    If ``suppress_print: False`` is part of the ``**kwargs`` of the
    wrapped method the output won't be suppressed.
    """

    @wraps(func)
    def wrapped(*args: Any, **kwargs: Any) -> Any:
        if not kwargs.pop("suppress_print", True):
            return func(*args, **kwargs)

        trap = io.StringIO()

        with contextlib.redirect_stdout(trap), contextlib.redirect_stderr(trap):
            result = func(*args, **kwargs)

        return result

    return wrapped


def merge_dicts(d1: Dict, d2: Dict) -> Dict:
    """
    Update two dicts of dicts recursively, if either mapping has leaves
    that are non-dicts, the second's leaf overwrites the first's.
    Taken from: https://stackoverflow.com/a/24088493/3511979
    """
    for k, v in d1.items():
        if k in d2:
            if all(isinstance(e, MutableMapping) for e in (v, d2[k])):
                d2[k] = merge_dicts(v, d2[k])
    d3 = d1.copy()
    d3.update(d2)

    return d3


@contextlib.contextmanager
def profile() -> None:
    """
    Profiler that operates as a context manager. Example usage:

    .. code-block:: python

        with profile():
            foo()
            bar()
    """
    pr = cProfile.Profile()
    pr.enable()

    yield

    pr.disable()
    pstats.Stats(pr).sort_stats("cumulative").print_stats()



import codecs  # To use a consistent encoding
import os
from os import path

from setuptools import find_packages, setup

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with codecs.open(path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

on_rtd = os.environ.get('READTHEDOCS') == 'True'
if on_rtd:
    install_requires = []
else:
    with open('requirements.txt') as f:
        install_requires = [l.strip() for l in f]

setup(
    name="diagNNose",
    version="1.1",
    description="A library that facilitates a broad set of tools for analysing "
                "hidden activations of neural models.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="Apache License 2.0",
    classifiers=[
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Programming Language :: Python :: 3.6",
    ],
    keywords="interpretability, analysis, deep learning, nlp",
    url="https://github.com/i-machine-think/diagnnose",
    packages=find_packages(exclude=["contrib", "docs", "test", "scripts"]),
    install_requires=install_requires,
    python_requires=">=3.6.0",
    extras_require={"dev": ["check-manifest"], "test": ["coverage"]},
)



from abc import ABC, abstractmethod
from typing import Optional, Union

import torch.nn as nn
from torch import Tensor

from diagnnose.attribute import ShapleyTensor
from diagnnose.typedefs.activations import (
    ActivationDict,
    ActivationName,
    ActivationNames,
    SizeDict,
)


class LanguageModel(ABC, nn.Module):
    device: str = "cpu"
    sizes: SizeDict = {}
    is_recurrent: bool = False

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @abstractmethod
    def forward(
        self,
        input_ids: Optional[Tensor] = None,
        inputs_embeds: Optional[Union[Tensor, ShapleyTensor]] = None,
        input_lengths: Optional[Tensor] = None,
        compute_out: bool = False,
        calc_causal_lm_probs: bool = False,
        only_return_top_embs: bool = False,
    ) -> Union[ActivationDict, Tensor]:
        """Performs a single forward pass across all LM layers.

        Parameters
        ----------
        input_ids : Tensor, optional
            Indices of input sequence tokens in the vocabulary.
            Size: batch_size x max_sen_len
        inputs_embeds : Tensor | ShapleyTensor, optional
            This is useful if you want more control over how to convert
            `input_ids` indices into associated vectors than the model's
            internal embedding lookup matrix. Also allows a
            ShapleyTensor to be provided, allowing feature contributions
            to be track during a forward pass.
            Size: batch_size x max_sen_len x nhid
        input_lengths : Tensor, optional
            Tensor of the sentence lengths of each batch item.
            If not provided, all items are assumed the same length.
            Size: batch_size,
        compute_out : bool, optional
            Toggles the computation of the final decoder projection.
            If set to False this projection is not calculated.
            Defaults to False.
        calc_causal_lm_probs : bool, optional
            Toggle to directly compute the output probabilities of the
            next token within the forward pass. Next token is inferred
            based on the shifted input tokens. Should only be toggled
            for Causal (a.k.a. auto-regressive) LMs. Defaults to False.
        only_return_top_embs : bool, optional
            Toggle to only return the tensor of the hidden state of the
            top layer of the network, instead of the full activation
            dictionary. Defaults to False.

        Returns
        -------
        activations : ActivationDict
            Dictionary mapping an activation name to a (padded) tensor.
            Size: a_name -> batch_size x max_sen_len x nhid
        """
        raise NotImplementedError

    @abstractmethod
    def create_inputs_embeds(self, input_ids: Tensor) -> Tensor:
        """Transforms a sequence of input tokens to their embedding.

        Parameters
        ----------
        input_ids : Tensor
            Tensor of shape batch_size x max_sen_len.

        Returns
        -------
        inputs_embeds : Tensor
            Embedded tokens of shape batch_size x max_sen_len x nhid.
        """
        raise NotImplementedError

    @abstractmethod
    def activation_names(self) -> ActivationNames:
        """Returns a list of all the model's activation names.

        Returns
        -------
        activation_names : ActivationNames
            List of (layer, name) tuples.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def num_layers(self) -> int:
        """ Returns the number of layers in the LM. """
        raise NotImplementedError

    @property
    @abstractmethod
    def top_layer(self) -> int:
        """ Returns the index of the LM's top layer. """
        raise NotImplementedError

    @abstractmethod
    def nhid(self, activation_name: ActivationName) -> int:
        """ Returns number of hidden units for a (layer, name) tuple. """
        raise NotImplementedError



from collections import namedtuple
from typing import Callable, Union

from torchtext.data import Example

DataDict = namedtuple(
    "DataDict",
    [
        "train_activations",
        "train_labels",
        "train_control_labels",
        "test_activations",
        "test_labels",
        "test_control_labels",
    ],
)

DataSplit = namedtuple("DataSplit", ["activation_reader", "labels", "control_labels"])

DCConfig = namedtuple(
    "DCConfig",
    ["lr", "max_epochs", "rank", "lambda1", "verbose"],
)

# https://www.aclweb.org/anthology/D19-1275/
# w position, batch item -> label
ControlTask = Callable[[int, Example], Union[str, int]]



from typing import Any, Dict

# Dict mapping: group -> arg_name -> arg_attr -> value
# Where arg_attr one of 'help', 'nargs', or 'type'.
arg_descriptions: Dict[str, Dict[str, Dict[str, Any]]] = {}

arg_descriptions["model"] = {
    # Gulordava/Forward LSTM
    "model_type": {"help": "RNN model type, as of now either ForwardLSTM or GoogleLM."},
    "state_dict": {"help": "Path to model parameters (pickled torch state_dict)"},
    "device": {
        "help": "(optional) Torch device name on which model will be run. Defaults to cpu."
    },
    "rnn_name": {"help": "(optional) Attribute name of rnn, defaults to `rnn`."},
    "encoder_name": {
        "help": "(optional) Attribute name of model encoder, defaults to `encoder`."
    },
    "decoder_name": {
        "help": "(optional) Attribute name of model decoder, defaults to `decoder`."
    },
    # Huggingface Transformer models
    "model_name": {"help": "Huggingface model name of pretrained LM."},
    # GoogleLM
    "pbtxt_path": {
        "help": "Path to the .pbtxt file containing the GraphDef model setup."
    },
    "ckpt_dir": {"help": "Path to folder containing parameter checkpoint files."},
    "full_vocab_path": {
        "help": "Path to the full model vocabulary of 800k tokens. Note that `vocab_path` "
        "can be passed along as well, pointing toward the corpus that will be extracted. "
        "In that case only a subset of the model softmax will be loaded in."
    },
    "corpus_vocab_path": {
        "help": "(optional) Path to the corpus for which a vocabulary will be created. "
        "This allows for only a subset of the model softmax to be loaded in."
    },
    "create_decoder": {
        "help": "(optional) Toggle to load in the (partial) softmax weights. Can be set to "
        "false in case no decoding projection needs to be made, as is the case during "
        "activation extraction, for example."
    },
}

arg_descriptions["tokenizer"] = {
    "path": {
        "help": "Path to the model vocabulary, which should a file containing a vocab "
        "entry at each line."
    },
    "notify_unk": {
        "type": bool,
        "help": "Notify when a requested token is not present in the vocab. Defaults to False.",
        "default": False,
    },
}

arg_descriptions["init_states"] = {
    "pickle_path": {
        "help": "(optional) Path to pickle of the initial lstm states of the model. "
        "If no path is provided zero-initialized states will be used."
    },
    "corpus_path": {
        "help": "(optional) Path to corpus for which the final activation will be used as "
        "initial states to the model."
    },
    "save_init_states_to": {
        "help": "(optional) Path to which the newly computed init_states will be saved. "
        "If not provided these states won't be dumped."
    },
    "use_default": {
        "type": bool,
        "help": "(optional) Toggle to use the default initial sentence `. <eos>`. "
        "Defaults to False",
        "default": False,
    },
}

arg_descriptions["corpus"] = {
    "path": {"help": "(required) Path to a corpus file."},
    "header": {
        "nargs": "*",
        "help": "(optional) List of corpus attribute names. If not provided all lines will be "
        'considered to be sentences, with the attribute name "sen".',
    },
    "to_lower": {
        "type": bool,
        "help": "(optional) Convert corpus to lowercase, defaults to False.",
        "default": False,
    },
    "header_from_first_line": {
        "type": bool,
        "help": "(optional) Use the first line of the corpus as the attribute  names of the "
        "corpus. Defaults to False.",
        "default": False,
    },
    "sep": {
        "help": "Column separator of corpus file, either a tsv or csv. Defaults to '\t'."
    },
}

arg_descriptions["extract"] = {
    "activations_dir": {
        "help": "Path to directory to which extracted embeddings will be written."
    },
    "activation_names": {
        "nargs": "*",
        "help": "List of activation names to be extracted. Activation names must "
        "be provided as a list of lists: i.e. [[layer, name]].",
    },
    "batch_size": {
        "type": int,
        "help": "(optional) Amount of sentences processed per forward step. "
        "Higher batch size increases extraction speed, but should "
        "be done accordingly to the amount of available RAM. Defaults to 1.",
    },
    "dtype": {
        "help": "(optional) Activation dtype, should be one of float32 or float64. "
        "Defaults to float32."
    },
}

arg_descriptions["init_dc"] = {
    "save_dir": {
        "help": "Directory to which trained models will be saved.",
    },
    "test_activations_dir": {
        "help": "(optional) Directory containing the extracted test activations. If not "
        "provided the train activation set will be split and partially used as test set."
    },
    "test_corpus": {
        "help": "(optional) Path to test corpus file, containing the test labels."
    },
    "classifier_type": {
        "help": "Either `logreg_torch`, using a torch logreg model, or  `logreg_sklearn`, "
        "using a LogisticRegressionCV model of sklearn."
    },
    "verbose": {"help": "Set to any positive number for verbosity. Defaults to 0."},
}

arg_descriptions["train_dc"] = {
    "calc_class_weights": {
        "type": bool,
        "help": "(optional) Set to true to calculate the classifier class weights based on the "
        "corpus class frequencies. Defaults to False.",
        "default": False,
    },
    "data_subset_size": {
        "type": int,
        "help": "(optional) Subset size of the amount of data points that will be used for "
        "training. Train/test split is performed afterwards. Defaults to the entire "
        "data set.",
    },
    "train_test_split": {
        "type": float,
        "help": "(optional) Ratio of the train/test split. Defaults to 0.9.",
    },
    "rank": {
        "type": int,
        "help": "Matrix rank of the linear classifier. Defaults to the full rank if not "
        "provided.",
    },
    "max_epochs": {
        "type": int,
        "help": "Maximum number of training epochs used by skorch. Defaults to 10.",
    },
}

arg_descriptions["decompose"] = {
    "decomposer": {
        "help": "Class name of decomposer constructor. As of now either "
        "CellDecomposer or ContextualDecomposer. Defaults to ContextualDecomposer."
    },
    "decompose_o": {
        "help": " Toggles decomposition of the output gate. Defaults to False."
    },
    "rel_interactions": {
        "nargs": "*",
        "help": " Indicates the interactions that are part of the relevant decomposition. "
        "Possible interactions are: rel-rel, rel-b and rel-irrel. Defaults to rel-rel, "
        "rel-irrel & rel-b.",
    },
    "bias_bias_only_in_phrase": {
        "help": " Toggles whether the bias-bias interaction should only be added when inside "
        "the relevant phrase. Defaults to True, indicating that only bias-bias "
        "interactions inside the subphrase range are added to the relevant decomposition."
    },
    "only_source_rel": {
        "help": " Relates to rel-irrel interactions. If set to true, only irrel_gate-rel_"
        "source interactions will be added to rel, similar to LRP (Arras et al., 2017)."
    },
    "only_source_rel_b": {
        "help": " Relates to rel-b interactions. If set to true, only b-rel_source "
        "interactions will be added to rel, similar to LRP (Arras et al., 2017)."
    },
    "input_never_rel": {
        "help": " Never add the Wx input to the rel part, useful when only investigating "
        "the model biases. Defaults to False."
    },
    "init_states_rel": {
        "help": " Directly add the initial cell/hidden states to the relevant part. "
        "Defaults to False."
    },
    "use_extracted_activations": {
        "help": " Allows previously extracted activations to be used to avoid unnecessary "
        "recomputations of those activations. Defaults to True."
    },
    "only_return_dec": {
        "help": " Only returns the decomposed cell states of the top layer, without "
        "calculating the corresponding decoder scores. Defaults to False."
    },
    "normalize": {
        "help": "Toggle to normalize the relevant scores by the full model logits."
    },
    "normalize_w_betas": {
        "help": "Toggle to normalize the relevant scores by the sum of relevant scores."
    },
}

arg_descriptions["plot_attention"] = {
    "title": {"help": "(optional) Title of the plot."},
    "clim": {
        "nargs": "*",
        "help": "(optional) Color limits. Should be be provided as [cmin, cmax], "
        "or [cmin, cmid, cmax].",
    },
    "cmap": {"help": "(optional) Pyplot colormap. Defaults to `coolwarm`."},
    "plot_values": {
        "help": "(optional) Toggle to plot the individual values. Defaults to True."
    },
    "value_font_size": {
        "help": "(optional) Font size of the individual plotted values."
    },
    "patches": {
        "nargs": "*",
        "help": "(optional) List of [x, y] tuples denoting highlighted patch indices.",
    },
}

arg_descriptions["downstream"] = {
    "config": {
        "help": "Dictionary mapping a downstream task to a config dict relating to that task."
        "Current supported tasks are `linzen`, `marvin`, `lakretz`, and `warstadt`."
    },
    "tasks": {
        "nargs": "*",
        "help": "List of tasks for which the downstream setup should be computed. This allows"
        "to store the config of each task in the json if one wishes to skip it momentarily.",
    },
}



import shutil
from typing import Optional, Tuple

from diagnnose.activations import ActivationReader
from diagnnose.activations.selection_funcs import return_all
from diagnnose.corpus import Corpus
from diagnnose.models import LanguageModel
from diagnnose.typedefs.activations import (
    ActivationNames,
    RemoveCallback,
    SelectionFunc,
)
from diagnnose.utils.misc import suppress_print

from .extractor import BATCH_SIZE, Extractor


@suppress_print
def simple_extract(
    model: LanguageModel,
    corpus: Corpus,
    activation_names: ActivationNames,
    activations_dir: Optional[str] = None,
    batch_size: int = BATCH_SIZE,
    selection_func: SelectionFunc = return_all,
) -> Tuple[ActivationReader, RemoveCallback]:
    """Basic extraction method.

    Parameters
    ----------
    model : LanguageModel
        Language model that inherits from LanguageModel.
    corpus : Corpus
        Corpus containing sentences to be extracted.
    activation_names : List[tuple[int, str]]
        List of (layer, activation_name) tuples
    activations_dir : str, optional
        Directory to which activations will be written. If not provided
        the `extract()` method will only return the activations without
        writing them to disk.
    selection_func : SelectionFunc
        Function which determines if activations for a token should
        be extracted or not.
    batch_size : int, optional
        Amount of sentences processed per forward step. Higher batch
        size increases extraction speed, but should be done
        accordingly to the amount of available RAM. Defaults to 1.

    Returns
    -------
    activation_reader : ActivationReader
        ActivationReader for the activations that have been extracted.
    remove_activations : RemoveCallback
        Callback function that can be executed at the end of a procedure
        that depends on the extracted activations. Removes all the
        activations that have been extracted. Takes no arguments.
    """
    extractor = Extractor(
        model,
        corpus,
        activation_names,
        activations_dir=activations_dir,
        selection_func=selection_func,
        batch_size=batch_size,
    )

    activation_reader = extractor.extract()

    def remove_activations():
        if activations_dir is not None:
            shutil.rmtree(activations_dir)

    return activation_reader, remove_activations



from typing import Type

from .init_states import _create_zero_states
from .language_model import LanguageModel
from .recurrent_lm import RecurrentLM
from .transformer_lm import TransformerLM


def import_model(*args, **kwargs) -> LanguageModel:
    """Import a model from a json file.

    Returns
    --------
    model : LanguageModel
        A LanguageModel instance, based on the provided config_dict.
    """

    if "transformer_type" in kwargs:
        model = _import_transformer_lm(*args, **kwargs)
    elif "rnn_type" in kwargs:
        model = _import_recurrent_lm(*args, **kwargs)
    else:
        raise TypeError("`transformer_type` or `rnn_type` must be provided as kwarg.")

    model.eval()

    device = kwargs.get("device", "cpu")
    model.device = device
    model.to(device)

    return model


def _import_transformer_lm(*args, **kwargs) -> TransformerLM:
    """ Imports a Transformer LM. """
    return TransformerLM(*args, **kwargs)


def _import_recurrent_lm(*args, **kwargs) -> RecurrentLM:
    """ Imports a recurrent LM and sets the initial states. """

    assert "rnn_type" in kwargs, "rnn_type should be given for recurrent LM"
    model_type = kwargs.pop("rnn_type")

    import diagnnose.models.wrappers as wrappers

    model_constructor: Type[RecurrentLM] = getattr(wrappers, model_type)
    model = model_constructor(*args, **kwargs)

    model.init_states = _create_zero_states(model)

    return model



from typing import Dict
from warnings import warn


class W2I(dict):
    """Provides vocab functionality mapping words to indices.

    Non-existing tokens are mapped to the id of an unk token that should
    be present in the vocab file.

    Parameters
    ----------
    w2i : Dict[str, int]
        Dictionary that maps strings to indices. This dictionary can be
        created using `create_vocab`.
    unk_token : str, optional
        The unk token to which unknown words will be mapped. Defaults to
        <unk>.
    eos_token : str, optional
        The end-of-sentence token that is used in the corpus. Defaults
        to <eos>.
    notify_unk : bool, optional
        Notify when a requested token is not present in the vocab.
        Defaults to False.
    """

    def __init__(
        self,
        w2i: Dict[str, int],
        unk_token: str = "<unk>",
        eos_token: str = "<eos>",
        pad_token: str = "<pad>",
        notify_unk: bool = False,
    ) -> None:
        super().__init__(w2i)

        if unk_token not in w2i:
            warn(f"Unk token {unk_token} not found in provided vocab.")

        self.unk_token = unk_token
        self.eos_token = eos_token
        self.pad_token = pad_token

        self.unk_idx = w2i[unk_token]

        self.notify_unk = notify_unk

        self.i2w = list(w2i.keys())

    @property
    def w2i(self) -> Dict[str, int]:
        return self

    def __missing__(self, key: str) -> int:
        if self.notify_unk:
            warn(f"`{key}` is not present in vocab")
        return self.unk_idx



import os
import pickle
from typing import Any

import dill


def load_pickle(path: str, use_dill: bool = False) -> Any:
    with open(path, "rb") as f:
        if use_dill:
            out = dill.load(f)
        else:
            out = pickle.load(f)
    return out


def dump_pickle(content: Any, path: str) -> None:
    with open(os.path.expanduser(path), "wb") as f:
        pickle.dump(content, f)



import os
import warnings
from typing import Dict, List, NamedTuple, Optional, Tuple

from torchtext.data import Example, Field
from unidecode import unidecode

from diagnnose.corpus import Corpus
from diagnnose.typedefs.syntax import SyntaxEvalCorpora

from ..task import SyntaxEvalTask


class RawItem(NamedTuple):
    """ The original corpus structure contains these 18 fields. """

    sentence: str
    orig_sentence: str
    pos_sentence: str
    subj: str
    verb: str
    subj_pos: str
    has_rel: str
    has_nsubj: str
    verb_pos: str
    subj_index: str
    verb_index: str
    n_intervening: str
    last_intervening: str
    n_diff_intervening: str
    distance: str
    max_depth: str
    all_nouns: str
    nouns_up_to_verb: str


class LinzenTask(SyntaxEvalTask):
    def initialize(
        self,
        path: str,
        subtasks: Optional[List[str]] = None,
        items_per_subtask: Optional[int] = 1000,
    ) -> SyntaxEvalCorpora:
        """Performs the initialization for the tasks of
        Linzen et al. (2016)

        Arxiv link: https://arxiv.org/abs/1611.01368

        Repo: https://github.com/TalLinzen/rnn_agreement

        Parameters
        ----------
        path : str
            Path to directory containing the Marvin datasets that can be
            found in the github repo.
        subtasks : List[str], optional
            The downstream tasks that will be tested. If not provided this
            will default to the full set of conditions.
        items_per_subtask : int, optional
            Number of items that is selected per subtask. If not
            provided the full subtask set will be used instead.

        Returns
        -------
        corpora : Dict[str, Corpus]
            Dictionary mapping a subtask to a Corpus.
        """
        subtasks = subtasks or ["SS", "SP", "PS", "PP", "SPP", "PSS", "SPPP", "PSSS"]

        corpora: SyntaxEvalCorpora = self._create_corpora(
            path, subtasks, items_per_subtask
        )

        return corpora

    def _create_corpora(
        self, corpus_path: str, subtasks: List[str], items_per_subtask: Optional[int]
    ) -> SyntaxEvalCorpora:
        raw_corpora: Dict[str, List[RawItem]] = self._create_raw_corpora(
            corpus_path, subtasks
        )

        corpora = {}
        verb_inflections = self._create_verb_inflections(corpus_path)

        for condition, items in raw_corpora.items():
            corpus = self._create_corpus(items, verb_inflections, items_per_subtask)

            n_attractors = str(len(condition))
            corpora.setdefault(n_attractors, {})[condition] = corpus

        return corpora

    def _create_raw_corpora(self, corpus_path, subtasks) -> Dict[str, List[RawItem]]:
        raw_corpora = {}

        with open(os.path.join(corpus_path, "agr_50_mostcommon_10K.tsv")) as f:
            next(f)  # skip header
            for line in f:
                item: RawItem = RawItem(*line.strip().split("\t"))
                sva_condition = self._item_to_sva_condition(item)
                if self._item_to_sva_condition(item) in subtasks:
                    raw_corpora.setdefault(sva_condition, []).append(item)

        return raw_corpora

    def _create_corpus(
        self,
        items: List[RawItem],
        verb_inflections: Dict[str, str],
        items_per_subtask: Optional[int],
    ) -> Corpus:
        header = ["sen", "token", "counter_token"]
        fields = Corpus.create_fields(header, tokenizer=self.tokenizer)

        examples: List[Optional[Example]] = [
            self._item_to_example(item, fields, verb_inflections) for item in items
        ]

        examples: List[Example] = list(filter(None, examples))

        if items_per_subtask is not None:
            examples = examples[:items_per_subtask]

        corpus = Corpus(examples, fields)

        return corpus

    @staticmethod
    def _item_to_sva_condition(item: RawItem) -> str:
        """Maps an item to an SVA condition, based on the POS tags of
        the sentence between the subject and the main verb.

        For example:
        The_DT men_NNS under_IN the_DT bridge_NN walk_VBP
        is mapped to "PS": a plural subject with 1 singular attractor.
        """
        pos_sen = item.pos_sentence.split()
        pos_subsen = pos_sen[int(item.subj_index) - 1 : int(item.verb_index)]

        pos_mapping = {"NNS": "P", "NN": "S"}
        sva_condition = "".join([pos_mapping.get(t, "") for t in pos_subsen])

        return sva_condition

    @staticmethod
    def _item_to_example(
        item: RawItem, fields: List[Tuple[str, Field]], verb_inflections: Dict[str, str]
    ) -> Optional[Example]:
        """Creates an Example containing the subsentence and both
        forms of the verb. If a verb form is not present in the model
        vocab, None is returned.
        """
        orig_sentence = item.orig_sentence

        if not all(ord(c) < 256 for c in orig_sentence):
            orig_sentence = unidecode(orig_sentence)

        subsen = orig_sentence.split()[: int(item.verb_index) - 1]
        opposite_verb = verb_inflections.get(item.verb, None)

        if opposite_verb is None:
            return None

        return Example.fromlist([subsen, item.verb, opposite_verb], fields)

    def _create_verb_inflections(self, corpus_path: str) -> Dict[str, str]:
        """Create sing<>plur mappings for all verbs in the model vocab.

        Mappings are based on the pos-tagged wiki.vocab file of Linzen
        et al., and the `inflect` library.

        We only add token mappings if both the singular and plural form
        is present in the model's vocabulary.
        """
        try:
            import inflect
        except ImportError:
            warnings.warn("`inflect` is needed, can be installed using pip")
            raise

        infl_eng = inflect.engine()

        pos_to_token = {"VBP": [], "VBZ": []}

        try:
            with open(os.path.join(corpus_path, "wiki.vocab")) as file:
                next(file)
                for line in file:
                    word, pos, _ = line.strip().split()
                    if word in self.tokenizer.vocab and pos in pos_to_token:
                        pos_to_token[pos].append(word)
        except FileNotFoundError:
            warnings.warn(
                "wiki.vocab is expected to be located in the same directory as the full corpus."
            )
            raise

        verb_inflections = {}
        for word in pos_to_token["VBZ"]:
            candidate = infl_eng.plural_verb(word)
            if candidate in pos_to_token["VBP"]:
                verb_inflections[candidate] = word
                verb_inflections[word] = candidate

        return verb_inflections
